# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2015, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

__all__ = ['Config', 'VerbConfig', 'AttributeFilter', 'ParameterSet']

import inspect
import string

import prestans.types
import prestans.http
import prestans.exception               

class ParameterSet(object):
    """
    ParameterSet is a group of Dataprestans.types that are expected as GET parameters

    ParameterSet defines rules and patterns in which they are acceptable.
    while ParserRuleSet is responsible for running the parse mechanism.
    """

    def blueprint(self):
        """
        blueprint support, returns a partial dictionary
        """

        blueprint = dict()
        blueprint['type'] = "%s.%s" % (self.__module__, self.__class__.__name__)

        # Fields
        fields = dict()
        model_class_members = inspect.getmembers(self.__class__)
    
        # Inspects the attributes of a parameter set and tries to validate the input 
        for attribute_name, type_instance in self.__class__.__dict__.iteritems():
            
            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods 
                continue

            # Must be one of the following types
            if not isinstance(type_instance, prestans.types.String) and \
            not isinstance(type_instance, prestans.types.Float) and \
            not isinstance(type_instance, prestans.types.Integer) and \
            not isinstance(type_instance, prestans.types.Date) and \
            not isinstance(type_instance, prestans.types.DateTime) and \
            not isinstance(type_instance, prestans.types.Array):
                raise TypeError("%s should be subclass of prestans.types.String/Integer/Float/Date/DateTime/Array" % attribute_name)

            if isinstance(type_instance, prestans.types.Array):
                if not isinstance(type_instance.element_template, prestans.types.String) and \
                not isinstance(type_instance.element_template, prestans.types.Float) and \
                not isinstance(type_instance.element_template, prestans.types.Integer):
                    raise TypeError("%s should be subclass of prestans.types.String/Integer/Float/Array" % attribute_name)

            fields[attribute_name] = type_instance.blueprint()

        blueprint['fields'] = fields
        return blueprint
    
    def validate(self, request):
        """
        validate method for %ParameterSet
        
        Since the introduction of ResponseFieldListParser, the parameter _response_field_list 
        will be ignore, this is a prestans reserved parameter, and cannot be used by apps.
        
        @param self The object pointer
        @param request The request object to be validated
        """
        
        validated_parameter_set = self.__class__()

        # Inspects the attributes of a parameter set and tries to validate the input 
        for attribute_name, type_instance in self.__class__.__dict__.iteritems():
            
            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods 
                continue

            #: Must be one of the following types
            if not isinstance(type_instance, prestans.types.String) and \
            not isinstance(type_instance, prestans.types.Float) and \
            not isinstance(type_instance, prestans.types.Integer) and \
            not isinstance(type_instance, prestans.types.Date) and \
            not isinstance(type_instance, prestans.types.DateTime) and \
            not isinstance(type_instance, prestans.types.Array):
                raise TypeError("%s should be of type prestans.types.String/Integer/Float/Date/DateTime/Array" % attribute_name)

            if issubclass(type_instance.__class__, prestans.types.Array):
                
                if not isinstance(type_instance._element_template, prestans.types.String) and \
                not isinstance(type_instance._element_template, prestans.types.Float) and \
                not isinstance(type_instance._element_template, prestans.types.Integer):
                    raise TypeError("%s elements should be of type prestans.types.String/Integer/Float" % attribute_name)

            try:

                #: Get input from parameters
                #: Empty list returned if key is missing for getall
                if issubclass(type_instance.__class__, prestans.types.Array):
                    validation_input = request.params.getall(attribute_name)
                #: Key error thrown if key is missing for getone
                else:
                    try:
                        validation_input = request.params.getone(attribute_name)
                    except KeyError:
                        validation_input = None

                #: Validate input based on data type rules, raises DataTypeValidationException if validation fails 
                validation_result = type_instance.validate(validation_input)

                setattr(validated_parameter_set, attribute_name, validation_result)

            except prestans.exception.DataValidationException, exp:
                raise prestans.exception.ValidationError(
                    message=str(exp), 
                    attribute_name=attribute_name, 
                    value=validation_input, 
                    blueprint=type_instance.blueprint())
            
        return validated_parameter_set


class AttributeFilter(object):
    """     
      { 
        field_name0: true, 
        field_name1: false, 
        collection_name0: true, 
        collection_name1: false,
        collection_name2: {
            sub_field_name0: true,
            sub_field_name1: false 
        }
      }
    """

    @classmethod
    def from_model(self, model_instance, default_value=False, **kwargs):
        """
        wrapper for Model's get_attribute_filter
        """

        if not isinstance(model_instance, prestans.types.DataCollection):
            raise TypeError("model_instance must be a sublcass of prestans.types.DataCollection, %s given" % 
                            (model_instance.__class__.__name__))

        attribute_filter_instance = model_instance.get_attribute_filter(default_value)

        #: kwargs support
        for name, value in kwargs.iteritems():
            if attribute_filter_instance.__dict__.has_key(name):
                setattr(attribute_filter_instance, name, value)
            else:
                raise KeyError(name)

        return attribute_filter_instance            

    def __init__(self, from_dictionary=None, template_model=None, **kwargs):
        """
        Creates an attribute filter object, optionally populates from a 
        dictionary of booleans
        """

        if from_dictionary:
            self._init_from_dictionary(from_dictionary, template_model)

        #: kwargs support
        for name, value in kwargs.iteritems():
            if self.__dict__.has_key(name):
                setattr(self, name, value)
            else:
                raise KeyError(name)


    def conforms_to_template_filter(self, template_filter):
        """
        Check AttributeFilter conforms to the rules set by the template

         - If self, has attributes that template_fitler does not contain, throw Exception
         - If sub list found, perform the first check
         - If self has a value for an attribute, assign to final AttributeFilter
         - If not found, assign value from template
                  
        """     

        if not isinstance(template_filter, self.__class__):
            raise TypeError("AttributeFilter can only check conformance against another template filter, %s provided" % 
                            (template_filter.__class__.__name__))

        #:
        #: Keys from the template
        #:
        template_filter_keys = template_filter.keys()
        # Keys from the object itself
        this_filter_keys = self.keys()

        #:
        #: 1. Check to see if the client has provided unwanted keys
        #:
        unwanted_keys = set(this_filter_keys) - set(template_filter_keys)
        if len(unwanted_keys) > 0:
            raise prestans.exception.AttributeFilterDiffers(list(unwanted_keys))

        #:
        #: 2. Make a attribute_filter that we send back
        #:
        evaluated_attribute_filter = AttributeFilter()

        #:
        #: 3. Evaluate the differences between the two, with template_filter as the standard
        #:
        for template_key in template_filter_keys:

            if template_key in this_filter_keys:

                value = getattr(self, template_key)

                #:
                #: If sub filter and boolean provided with of true, create default filter 
                #: with value of true
                #:
                if isinstance(value, bool) and \
                value is True and \
                isinstance(getattr(template_filter, template_key), AttributeFilter):
                    setattr(evaluated_attribute_filter, template_key, getattr(template_filter, template_key))
                elif isinstance(value, bool):
                    setattr(evaluated_attribute_filter, template_key, value)
                elif isinstance(value, self.__class__):
                    # Attribute lists sort themselves out, to produce sub Attribute Filters 
                    template_sub_list = getattr(template_filter, template_key)
                    this_sub_list = getattr(self, template_key)
                    setattr(evaluated_attribute_filter, template_key, this_sub_list.conforms_to_template_filter(template_sub_list))
            else:
                setattr(evaluated_attribute_filter, template_key, getattr(template_filter, template_key))

        return evaluated_attribute_filter
        
    def keys(self):
        """ 
        returns a list of usable keys
        """ 

        keys = list()

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods
                continue

            keys.append(attribute_name)

        return keys

    def has_key(self, key):
        """
        contains a particular key, wrapper on self.__dict__.key
        """ 
        return self.__dict__.has_key(key)

    def is_filter_at_key(self, key):
        """ 
        return True if attribute is a subfilter
        """

        if self.has_key(key) and isinstance(attribute_status, self.__class__):
            return True

        return False

    def is_attribute_visible(self, key):
        """
        returns True if an attribute is visible
        
        If attribute is an instance of AttributeFilter, it returns True if all attributes
        of the sub filter are visible.
        """         
        if self.has_key(key):
            attribute_status = getattr(self, key)
            if isinstance(attribute_status, bool) and attribute_status == True:
                return True
            elif isinstance(attribute_status, self.__class__) and \
            attribute_status.are_any_attributes_visible():
                return True

        return False

    def are_any_attributes_visible(self):
        """
        checks to see if any attributes are set to true
        """ 

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods 
                continue

            if isinstance(type_instance, bool) and type_instance == True:
                return True
            elif isinstance(type_instance, self.__class__) and \
            type_instance.are_all_attributes_visible() == True:
                # Serialise attribute filter children to dictioanaries 
                return True

        return False

    def are_all_attributes_visible(self):
        """
        checks to see if all attributes are set to true
        """         

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods
                continue

            if isinstance(type_instance, bool) and type_instance == False:
                return False
            elif isinstance(type_instance, self.__class__) and type_instance.are_all_attributes_visible() == False:
                # Serialise attribute filter children to dictioanaries 
                return False

        return True

    def set_all_attribute_values(self, value):
        """
        sets all the attribute values to the value and propagate to any children
        """         

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods
                continue

            if isinstance(type_instance, bool):
                self.__dict__[attribute_name] = value
            elif isinstance(type_instance, self.__class__):
                # Serialise attribute filter children to dictioanaries 
                type_instance.set_all_attribute_values(value)


    def as_dict(self):
        """ 
        turns attribute filter object into python dictionary
        """

        output_dictionary = dict()

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods 
                continue

            if isinstance(type_instance, bool):
                output_dictionary[attribute_name] = type_instance
            elif isinstance(type_instance, self.__class__):
                # Serialise attribute filter children to dictioanaries 
                output_dictionary[attribute_name] = type_instance.as_dict()

        return output_dictionary


    def _init_from_dictionary(self, from_dictionary, template_model=None):
        """ 
        Private helper to init values from a dictionary, wraps chidlren into 
        AttributeFilter objects
        """

        if not isinstance(from_dictionary, dict):
            raise TypeError("from_dictionary must be of type dict, %s provided" % 
                            (from_dictionary.__class__.__name__))
        rewrite_map = None
        if template_model is not None:

            rewrite_map = template_model.attribute_rewrite_reverse_map()

            if not isinstance(template_model, prestans.types.DataCollection):
                raise TypeError("template_model should be a prestans model in AttributeFilter init (from dictionary), %s provided" % 
                                (template_model.__class__.__name__))

        for key, value in from_dictionary.iteritems():

            target_key = key

            #:
            #: Minification support
            #:
            if rewrite_map is not None:
                target_key = rewrite_map[key]

            #:
            #: Check to see we can work with the value 
            #:
            if not isinstance(value, (bool, dict)):
                raise TypeError("AttributeFilter input for key %s must be boolean or dict, %s provided" % 
                                (key, value.__class__.__name__))

            #:
            #: Ensure that the key exists in the template model
            #:
            if template_model is not None and not template_model.has_key(target_key):

                unwanted_keys = list()
                unwanted_keys.append(target_key)
                raise prestans.exception.AttributeFilterDiffers(unwanted_keys)

            #:
            #: Either keep the value of wrap it up with AttributeFilter 
            #:
            if isinstance(value, bool):
                setattr(self, target_key, value)
            elif isinstance(value, dict):

                sub_map = None
                if template_model is not None:

                    sub_map = getattr(template_model, target_key)

                    #: prestans Array support
                    if isinstance(sub_map, prestans.types.Array):
                        sub_map = sub_map.element_template

                setattr(self, target_key, AttributeFilter(from_dictionary=value, 
                    template_model=sub_map))
            

    def __setattr__(self, key, value):
        """
        Overrides setattr to allow only booleans or an AttributeFilter
        """ 

        # Set internal fields
        if key[0:1] == "_":
            self.__dict__[key] = value
            return

        # Values should either be boolean or type of self
        if isinstance(value, bool) and key in self.__dict__ and isinstance(self.__dict__[key], self.__class__):
            self.__dict__[key].set_all_attribute_values(value)
            return
        elif isinstance(value, (bool, self.__class__)):
            self.__dict__[key] = value
            return

        raise TypeError("%s name in %s must be of type Boolean or AttributeFilter, given %s" % 
                        (key, self.__class__.__name__, value.__class__.__name__))


class VerbConfig(object):
    """
    VerbConfig sets out rules for each HTTP Verb that your API will make available.
    These rules are used by prestans to validate requests and responses before
    handing over execution control to your handler.

    All verbs in use must provide atleast a response_template which should be a
    subclass of prestans.types.DataCollection.
    """

    def __init__(self, response_template=None, response_attribute_filter_default_value=False,
        parameter_sets=[], body_template=None, request_attribute_filter=None):

        #: response_template; required parameter        
        if response_template is not None and (not isinstance(response_template, prestans.types.DataCollection) and\
         not isinstance(response_template, prestans.types.BinaryResponse)):
            raise TypeError(
            "response_template of type %s must be an instance of a prestans.types.DataCollection subclass" % 
            response_template.__class__.__name__)

        if response_template is not None and isinstance(response_template, prestans.types.DataCollection):
            self._response_attribute_filter_template = AttributeFilter.from_model(model_instance=response_template, 
                default_value=response_attribute_filter_default_value)
        else:
            self._response_attribute_filter_template = None

        self._response_template = response_template


        #: parameter_sets
        if not isinstance(parameter_sets, list):
            parameter_sets = [parameter_sets]

        for parameter_set in parameter_sets:
            if not isinstance(parameter_set, ParameterSet):
                raise TypeError(
                "parameter_set of type %s must be an instance of prestans.parser.ParameterSet" %
                parameter_set.__class__.__name__)

        self._parameter_sets = parameter_sets

        #: body_template
        if body_template is not None and not isinstance(body_template, prestans.types.DataCollection):
            raise TypeError(
                "body_template of type %s must be an instance of a prestans.types.DataCollection subclass" %
                body_template.__class__.__name__)

        self._body_template = body_template

        #: request_attribute_filter
        if request_attribute_filter is not None and not isinstance(request_attribute_filter, AttributeFilter):
            raise TypeError(
            "request_attribute_filter of type %s must an instance of prestans.parser.AttributeFilter" % 
            request_attribute_filter.__class__.__name__)

        self._request_attribute_filter = request_attribute_filter

    def blueprint(self):
        
        verb_config_blueprint = dict()

        if self._response_template is not None:
            verb_config_blueprint['response_template'] = self._response_template.blueprint()
        else:
            verb_config_blueprint['response_template'] = self._response_template

        verb_config_blueprint['parameter_sets'] = []
        for parameter_set in self._parameter_sets:
            if parameter_set is not None:
                verb_config_blueprint['parameter_sets'].append(parameter_set.blueprint())
        
        if self._body_template is not None:
            verb_config_blueprint['body_template'] = self._body_template.blueprint()
        else:
            verb_config_blueprint['body_template'] = self._body_template

        if self._request_attribute_filter is not None:
            verb_config_blueprint['request_attribute_filter'] = self._request_attribute_filter.blueprint()
        else:
            verb_config_blueprint['request_attribute_filter'] = self._request_attribute_filter

        return verb_config_blueprint

    @property
    def response_template(self):
        return self._response_template

    @property 
    def response_attribute_filter_template(self):
        return self._response_attribute_filter_template

    @property
    def parameter_sets(self):
        return self._parameter_sets

    @property
    def body_template(self):
        return self._body_template

    @property
    def request_attribute_filter(self):
        return self._request_attribute_filter

class Config(object):
    """
    Configuration that's attached to each handler to define rules for each 
    HTTP Verb. All HTTP verbs in use must have a configuration defined.

    __init__ takes in a VerbConfig instance for each parameter with names
    adjacent to the HTTP verb.
    """

    def __init__(self, GET=None, HEAD=None, POST=None, PUT=None, PATCH=None, DELETE=None):

        for verb in [GET, HEAD, POST, PUT, PATCH, DELETE]:
            if verb is not None and not isinstance(verb, VerbConfig):
                raise TypeError("All Parser configs should be of type prestans.parser.VerbConfig")

        self._configs = dict()

        self._configs[prestans.http.VERB.GET] = GET
        self._configs[prestans.http.VERB.HEAD] = HEAD
        self._configs[prestans.http.VERB.POST] = POST
        self._configs[prestans.http.VERB.PUT] = PUT
        self._configs[prestans.http.VERB.PATCH] = PATCH
        self._configs[prestans.http.VERB.DELETE] = DELETE

    def get_config_for_verb(self, verb):
        return self._configs[verb]

    @property
    def GET(self):
        return self._configs[prestans.http.VERB.GET]

    @property
    def HEAD(self):
        return self._configs[prestans.http.VERB.HEAD]

    @property
    def POST(self):
        return self._configs[prestans.http.VERB.POST]

    @property
    def PUT(self):
        return self._configs[prestans.http.VERB.PUT]

    @property
    def PATCH(self):
        return self._configs[prestans.http.VERB.PATCH]

    @property
    def DELETE(self):
        return self._configs[prestans.http.VERB.DELETE]
     


# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2015, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import os
import re

import prestans.devel.gen
import prestans.types

def udl_to_cc(text, ignoreFirst=False):
    text = text.lower()
    camel_case = re.sub(r"_(.)", lambda pat: pat.group(1).upper(), text)
    if not ignoreFirst:
        camel_case = camel_case[0:1].upper()+camel_case[1:]
    return camel_case

class BasicTypeElementTemplate(object):

    def __init__(self, blueprint_type, blueprint):
        
        self._blueprint_type = blueprint_type
        self._required = None
        self._default = None
        self._minimum = None
        self._maximum = None
        self._min_length = None
        self._max_length = None
        self._choices = None
        self._format = None

        if self._blueprint_type == "string":
            self._required = blueprint['required']
            self._min_length = blueprint['min_length']
            self._max_length = blueprint['max_length']
            self._default = blueprint['default']
            self._choices = blueprint['choices']
            self._format = blueprint['format']
            self._client_class_name = "String"
        elif self._blueprint_type == 'integer':
            self._required = blueprint['required']
            self._default = blueprint['default']
            self._minimum = blueprint['minimum']
            self._maximum = blueprint['maximum']
            self._choices = blueprint['choices']
            self._client_class_name = "Integer"
        elif self._blueprint_type == 'float':
            self._required = blueprint['required']
            self._default = blueprint['default']
            self._minimum = blueprint['minimum']
            self._maximum = blueprint['maximum']
            self._choices = blueprint['choices']
            self._client_class_name = "Float"
        elif self._blueprint_type == 'boolean':
            self._required = blueprint['required']
            self._default = blueprint['default']
            self._client_class_name = "Boolean"

        if self._required is None:
            self._required = True

    @property
    def blueprint_type(self):
        return self._blueprint_type

    @property
    def client_class_name(self):
        return self._client_class_name

    @property
    def required(self):
        if self._required:
            return "true"
        else:
            return "false"

    @property
    def default(self):
        #dates are check first otherwise string will catch them
        #if self._default == prestans.types.DateTime.CONSTANT.NOW:
        #    return "prestans.types.DateTime.NOW"
        #elif self._default == prestans.types.Date.CONSTANT.TODAY:
        #    return "prestans.types.Date.TODAY"
        #elif self._default == prestans.types.Time.CONSTANT.NOW:
        #    return "prestans.types.Time.NOW"
        if self._default is None:
            return "null"
        elif type(self._default) == str:
            return "\"%s\"" % (self._default)
        elif type(self._default) == bool:
            if self._default:
                return "true"
            else:
                return "false"
        else:
            return self._default

    @property
    def minimum(self):
        if self._minimum is None:
            return "null"
        else:
            return self._minimum

    @property
    def maximum(self):
        if self._maximum is None:
            return "null"
        else:
            return self._maximum

    #string and array
    @property
    def min_length(self):
        if self._min_length is None:
            return "null"
        else:
            return self._min_length

    @property
    def max_length(self):
        if self._max_length is None:
            return "null"
        else:
            return self._max_length

    @property
    def format(self):
        if self._format is None:
            return "null"

        format = self._format.replace("\\", "\\\\")
        return "\"%s\"" % (format)

    @property
    def choices(self):
        if self._choices is None:
            return "null"
        else:
            return self._choices

class AttributeMetaData(object):

    def __init__(self, name, blueprint):
        self._name = name
        self._blueprint = blueprint
        self._required = None
        self._default = None
        self._minimum = None
        self._maximum = None
        self._min_length = None
        self._max_length = None
        self._choices = None
        self._format = None

        self._blueprint_type = blueprint['type']
        self._map_name = blueprint['map_name']

        #Basic types
        if self._blueprint_type == 'string':
            self._required = blueprint['constraints']['required']
            self._min_length = blueprint['constraints']['min_length']
            self._max_length = blueprint['constraints']['max_length']
            self._default = blueprint['constraints']['default']
            self._choices = blueprint['constraints']['choices']
            self._format = blueprint['constraints']['format']
            self._client_class_name = "String"

        elif self._blueprint_type == 'integer':
            self._required = blueprint['constraints']['required']
            self._default = blueprint['constraints']['default']
            self._minimum = blueprint['constraints']['minimum']
            self._maximum = blueprint['constraints']['maximum']
            self._choices = blueprint['constraints']['choices']
            self._client_class_name = "Integer"
        elif self._blueprint_type == 'float':
            self._required = blueprint['constraints']['required']
            self._default = blueprint['constraints']['default']
            self._minimum = blueprint['constraints']['minimum']
            self._maximum = blueprint['constraints']['maximum']
            self._choices = blueprint['constraints']['choices']
            self._client_class_name = "Float"
        elif self._blueprint_type == 'boolean':
            self._required = blueprint['constraints']['required']
            self._default = blueprint['constraints']['default']
            self._client_class_name = "Boolean"
        elif self._blueprint_type == 'datetime':
            self._required = blueprint['constraints']['required']
            self._default = blueprint['constraints']['default']
            self._timezone = blueprint['constraints']['timezone']
            self._utc = blueprint['constraints']['utc']
            self._client_class_name = "DateTime"
        elif self._blueprint_type == 'date':
            self._required = blueprint['constraints']['required']
            self._default = blueprint['constraints']['default']
            self._client_class_name = "Date"
        elif self._blueprint_type == 'time':
            self._required = blueprint['constraints']['required']
            self._default = blueprint['constraints']['default']
            self._client_class_name = "Time"
        elif self._blueprint_type == 'data_url_file':
            self._required = blueprint['constraints']['required']
            self._allowed_mime_types = blueprint['constraints']['allowed_mime_types']
            self._client_class_name = "DataURLFile"
        #Complex types
        elif self._blueprint_type == 'model':
            self._required = blueprint['constraints']['required']
            self._model_template = blueprint['constraints']['model_template']
            self._client_class_name = "Model"
        elif self._blueprint_type == 'array':
            self._required = blueprint['constraints']['required']
            self._min_length = blueprint['constraints']['min_length']
            self._max_length = blueprint['constraints']['max_length']
            self._client_class_name = "Array"

            element_template = blueprint['constraints']['element_template']

            if element_template['type'] == 'model':
                self._element_template_is_model = True
                self._element_template = element_template['constraints']['model_template']
            else:
                self._element_template_is_model = False
                self._element_template = BasicTypeElementTemplate(blueprint_type=element_template['type'], blueprint=element_template['constraints'])


    @property
    def name(self):
        return self._name
    
    @property
    def cc(self):
        return udl_to_cc(self._name)

    @property
    def ccif(self):
        return udl_to_cc(self._name, True)

    @property
    def blueprint_type(self):
        return self._blueprint_type

    @property
    def client_class_name(self):
        return self._client_class_name

    @property
    def map_name(self):
        return self._map_name

    @property
    def required(self):
        if self._required:
            return "true"
        else:
            return "false"

    @property
    def timezone(self):
        if self._timezone:
            return "true"
        else:
            return "false"

    @property
    def utc(self):
        if self._utc:
            return "true"
        else:
            return "false"

    @property
    def allowed_mime_types(self):
        return self._allowed_mime_types

    @property
    def default(self):
        #dates are check first otherwise string will catch them
        if self._default == prestans.types.DateTime.CONSTANT.NOW:
            return "prestans.types.DateTime.NOW"
        elif self._default == prestans.types.Date.CONSTANT.TODAY:
            return "prestans.types.Date.TODAY"
        elif self._default == prestans.types.Time.CONSTANT.NOW:
            return "prestans.types.Time.NOW"
        elif self._default is None:
            return "null"
        elif type(self._default) == str:
            return "\"%s\"" % (self._default)
        elif type(self._default) == bool:
            if self._default:
                return "true"
            else:
                return "false"
        else:
            return self._default

    @property
    def format(self):
        if self._format is None:
            return "null"

        format = self._format.replace("\\", "\\\\")
        return "\"%s\"" % (format)

    @property
    def minimum(self):
        if self._minimum is None:
            return "null"
        else:
            return self._minimum

    @property
    def maximum(self):
        if self._maximum is None:
            return "null"
        else:
            return self._maximum

    @property
    def choices(self):
        if self._choices is None:
            return "null"
        else:
            return self._choices

    #string and array
    @property
    def min_length(self):
        if self._min_length is None:
            return "null"
        else:
            return self._min_length

    @property
    def max_length(self):
        if self._max_length is None:
            return "null"
        else:
            return self._max_length

    #model
    @property
    def model_template(self):
        return self._model_template

    #array
    @property
    def element_template(self):
        return self._element_template

    @property
    def element_template_is_model(self):
        return self._element_template_is_model

class Base(object):

    def __init__(self, template_engine, model_file, namespace, output_directory):
        self._template_engine = template_engine
        self._model_file = model_file
        self._namespace = namespace
        self._output_directory = output_directory
        self._dependencies = list()
        self._attribute_string = ""

    def add_filter_dependency(self, attribute):

        dependency = None
        if attribute.blueprint_type == 'array' and attribute.element_template_is_model:
            dependency = "%s.%s" % (self._namespace, attribute.element_template)
        elif attribute.blueprint_type == 'model':
            dependency = "%s.%s" % (self._namespace, attribute.model_template)

        if dependency is not None and dependency not in self._dependencies:
            self._dependencies.append(dependency)

    def add_model_dependency(self, attribute):

        dependency = None
        if attribute.blueprint_type == 'array' and attribute.element_template_is_model:
            dependency = "%s.%s" % (self._namespace, attribute.element_template)
        elif attribute.blueprint_type == 'array':
            dependency = "%s.%s" % ("prestans.types", attribute.element_template.client_class_name)
        elif attribute.blueprint_type == 'model':
            dependency = "%s.%s" % (self._namespace, attribute.model_template)
        else:
            dependency = "prestans.types.%s" % (attribute.client_class_name)

        if dependency is not None and dependency not in self._dependencies:
            self._dependencies.append(dependency)

    #used in filters
    def add_attribute_string(self, attribute):
        if attribute.blueprint_type == 'model':
            self._attribute_string += "this.%s_.anyFieldsEnabled() || " % (attribute.ccif)
        elif attribute.blueprint_type == 'array' and attribute.element_template_is_model:
            self._attribute_string += "this.%s_.anyFieldsEnabled() || " % (attribute.ccif)
        else:
            self._attribute_string += "this.%s_ || " % (attribute.ccif)

    @property
    def attribute_string(self):
        return self._attribute_string[:-4]

class Model(Base):

    def __init__(self, template_engine, model_file, namespace, filter_namespace, output_directory):
    
        Base.__init__(self, template_engine, model_file, namespace, output_directory)
        self._filter_namespace = filter_namespace
        self._template = self._template_engine.get_template("closure/model.jinja")

    def run(self):

        inspector = prestans.devel.gen.Inspector(model_file=self._model_file)
        blueprints = inspector.inspect()

        for model_blueprint in blueprints:

            model_name = model_blueprint['constraints']['model_template']
            attributes = list()
            self._dependencies = list()
            self._attribute_string = ""

            for field_name, field_blueprint in model_blueprint['fields'].iteritems():

                attribute = AttributeMetaData(name=field_name, blueprint=field_blueprint)                
                attributes.append(attribute)

                self.add_model_dependency(attribute)

            #write out template
            filename = '%s.js' % (model_name)
            output_file = open(os.path.join(self._output_directory, filename), 'w+')

            output_file.write(self._template.render(namespace=self._namespace, filter_namespace=self._filter_namespace, name=model_name, attributes=attributes, dependencies=self._dependencies))
            output_file.close()

            print "%-30s -> %s.%s.js" %(model_name, self._namespace, model_name)

        print "\nGenerated %i model(s)" % len(blueprints)

        return 0

class Filter(Base):

    def __init__(self, template_engine, model_file, namespace, output_directory):

        Base.__init__(self, template_engine, model_file, namespace, output_directory)
        self._template = self._template_engine.get_template("closure/filter.jinja")

    def run(self):

        inspector = prestans.devel.gen.Inspector(model_file=self._model_file)
        blueprints = inspector.inspect()

        for model_blueprint in blueprints:

            model_name = model_blueprint['constraints']['model_template']
            attributes = list()
            self._dependencies = list()
            self._attribute_string = ""

            for field_name, field_blueprint in model_blueprint['fields'].iteritems():

                attribute = AttributeMetaData(name=field_name, blueprint=field_blueprint)
                attributes.append(attribute)

                self.add_filter_dependency(attribute)
                self.add_attribute_string(attribute)

            #write out template
            filename = '%s.js' % (model_name)
            output_file = open(os.path.join(self._output_directory, filename), 'w+')

            output_file.write(self._template.render(namespace=self._namespace, name=model_name, attributes=attributes, dependencies=self._dependencies, attribute_string=self.attribute_string))
            output_file.close()

            print "%-30s -> %s.%s.js" % (model_name, self._namespace, model_name)

        print "\nGenerated %i filter(s)" % len(blueprints)

        return 0



#!/usr/bin/env python
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2015, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from setuptools import setup, find_packages

setup(name='prestans',
      version='2.0.7',
      description='A WSGI compliant REST micro-framework',
      url='https://github.com/anomaly/prestans.git',
      long_description="""
      Prestans is a REST micro-framework built right on top of WSGI, designed 
      perform and co-exists with other middleware and frameworks that you employ. 
      Our target audience are developers building large REST backends for use 
      with pure Ajax (using a framework like Google Closure) or mobile applications. 
      Prestans is designed for you to "take as much or as little" you like, 
      although we do think that it all works really well together.

      Reasons we think you will love Prestans:

      - Built right on top of WSGI, designed ground up to cooperate with other frameworks
      - Strong representative of REST philosophies leveraging HTTP headers and verbs
      - Support for multiple dialects, including formalised patterns for binary content 
      - Handlers maps HTTP verbs to implemented class methods, complimented  with custom 
        request parser and response writer (built on top of WebOb)
      - Unforgivingly strict parsing of requests and responses to ensure data integrity
      - Helper methods to ease transformation of persistent objects to REST responses
      - Pluggable integration to authentication

      """,
      download_url='https://github.com/anomaly/prestans/archive/2.0.7.tar.gz',
      license='New BSD',
      author='Anomaly Software',
      author_email='support@anomaly.net.au',
      maintainer='Anomaly Software',
      maintainer_email='support@anomaly.net.au',
      platforms=['any'],
      scripts=['prestans/bin/pride'],
      packages=find_packages(),
      classifiers=[
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Topic :: Internet :: WWW/HTTP',
            'Topic :: Internet :: WWW/HTTP :: WSGI',
            'Topic :: Software Development :: Libraries :: Application Frameworks',
            'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      install_requires=['webob==1.4',],
     )



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2015, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

__all__ = ['VERB', 'STATUS']

#:
#: This file mostly contains constants for HTTP
#:

class VERB:
    """ 
    Encapsulates HTTP Verbs supported by the prestans framework in accordance
    with the REST definition. Each prestans REST handler must supporto at least
    one HTTP verb.

    HEAD is similar to a GET but it does not return a response.
    """

    GET    = "GET"
    HEAD   = "HEAD"
    POST   = "POST"
    PUT    = "PUT"
    PATCH  = "PATCH"
    DELETE = "DELETE"

    @classmethod
    def is_supported_verb(cls, method):
        return method in [VERB.GET, VERB.HEAD, VERB.POST, VERB.PUT, VERB.PATCH, VERB.DELETE]

class STATUS:
    """
    The following is a selection of HTTP status codes that are recommended for use 
    by REST services. You are welcome to use other available status codes.
    
    http://en.wikipedia.org/wiki/List_of_HTTP_status_codes
    """

    #: Informational

    CONTINUE                    = 100
    SWITCHING_PROTOCOLS         = 101

    #: Success

    OK                          = 200
    CREATED                     = 201
    ACCEPTED                    = 202
    NON_AUTH_INFORMATION        = 203
    NO_CONTENT                  = 204
    RESET_CONTENT               = 205
    PARTIAL_CONTENT             = 206

    #: Redirection

    MULTIPLE_CHOICES            = 300
    MOVED_PERMANENTLY           = 301
    FOUND                       = 302
    SEE_OTHER                   = 303
    NOT_MODIFIED                = 304
    USE_PROXY                   = 305
    SWITCH_PROXY                = 306
    TEMPORARY_REDIRECT          = 307
    PERMANENT_REDIRECT          = 308

    #: Client Error

    BAD_REQUEST                 = 400
    UNAUTHORIZED                = 401
    PAYMENT_REQUIRED            = 402
    FORBIDDEN                   = 403
    NOT_FOUND                   = 404
    METHOD_NOT_ALLOWED          = 405
    NOT_ACCEPTABLE              = 406
    PROXY_AUTH_REQUIRED         = 407
    REQUEST_TIMEOUT             = 408
    CONFLICT                    = 409
    GONE                        = 410
    LENGTH_REQUIRED             = 411
    PRECONDITION_FAILED         = 412
    REQUEST_ENTITY_TOO_LARGE    = 413
    REQUEST_URI_TOO_LARGE       = 414
    UNSUPPORTED_MEDIA_TYPE      = 415
    RANGE_NOT_SERIALIZABLE      = 416
    EXCEPTION_FAILED            = 417
    TEAPOT                      = 418
    AUTHENTICATION_TIMEOUT      = 419

    #: Server Error
    
    INTERNAL_SERVER_ERROR       = 500
    NOT_IMPLEMENTED             = 501
    BAD_GATEWAY                 = 502
    SERVICE_UNAVAILABLE         = 503
    GATEWAY_TIMEOUT             = 504
    UNSUPPORTED_HTTP_VERSION    = 505
    



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2015, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import os
import sys
import yaml
import importlib

import blessings

from voluptuous import Schema, Required, All, Length, Range, MultipleInvalid

from werkzeug.serving import run_simple
from werkzeug.wsgi import DispatcherMiddleware

import prestans.devel.exception

class Configuration:

    _SCHEMA = {
        Required('name'): str,
        Required('version'): float,
        Required('bind'): str,
        Required('port'): int,
        'append_path': [str],
        'environ': [{
            Required('key'): str,
            Required('value'): str
        }],
        'static': [{
            Required('url'): str,
            Required('path'): str
        }],
        Required('default_app'): str,
        Required('sub_modules'): All([{
            Required('url'): str,
            Required('module'): str,
        }], Length(min=1)),
    }

    def __init__(self, config_path):
        
        try:
            self.base_path = os.path.dirname(config_path)
            parsed_config = yaml.load(file(config_path))
        except IOError, exp:
            raise prestans.devel.exception.Base("[error] unable to read configuration at %s" % config_path)

        try:
            schema = Schema(Configuration._SCHEMA)
            validated_config = schema(parsed_config)
        except MultipleInvalid, exp:
            raise prestans.devel.exception.Base("[error/config] %s" % str(exp), 2)

        #: Make configuration vars into instance vars

        self.name = validated_config['name']
        self.version = validated_config['version']
        self.bind = validated_config['bind']
        self.port = validated_config['port']

        self.append_path = validated_config['append_path']
        self.environ = validated_config['environ']
        self.static = validated_config['static']
        self.default_app = validated_config['default_app']
        self.handlers = validated_config['sub_modules']

class DevServer(object):

    def __init__(self, config):
        self._config = config
        self._terminal = blessings.Terminal()

        #: Append paths
        self._append_paths()
        self._add_environment_vars()

    def run(self):

        _static_file_map = self._create_static_map()
        _handlers = self._add_handlers()

        #: handover to werkzeug
        run_simple(self._config.bind, self._config.port, _handlers,\
         static_files=_static_file_map, use_reloader=True, use_debugger=True, use_evalex=True)

    def _append_paths(self):    

        for path in self._config.append_path:

            module_path = os.path.join(self._config.base_path, path)

            if os.path.exists(module_path):
                sys.path.append(module_path)
            else:
                print "[{t.red}path{t.normal}] ignoring non-existant include path %s".format(t=self._terminal)\
                 % module_path

    def _add_environment_vars(self):

        for entry in self._config.environ:

            key = entry['key']
            value = entry['value']

            os.environ[key] = value

    def _create_static_map(self):

        static_map = dict()

        for entry in self._config.static:

            url = entry['url']
            path = entry['path']

            resource_path = os.path.join(self._config.base_path, path)
            static_map[url] = resource_path

        return static_map

    def _add_handlers(self):

        sub_maps = dict()
        default_application = None

        #: Default application
        default_module = self._config.default_app
        try:
            module_name, wsgi_app = default_module.rsplit(".", 1)
            imported_module = importlib.import_module(module_name)
            default_application = getattr(imported_module, wsgi_app)
        except (ImportError, AttributeError), exp:
            raise prestans.devel.exception.Base("[error] default_app module %s doesn't exists" % default_module)

        #: Sub modules
        for entry in self._config.handlers:
            
            url = entry['url']
            module = entry['module']

            module_name, wsgi_app = module.rsplit(".", 1)
            imported_module = importlib.import_module(module_name)

            try:
                sub_maps[url] = getattr(imported_module, wsgi_app)
            except AttributeError, exp:
                raise prestans.devel.exception.Base("[error] module %s doesn't exists" % entry)

        dispatched_application = DispatcherMiddleware(default_application, sub_maps)

        return dispatched_application
